import platform, logging
from enum import Enum
from binascii import b2a_hex, a2b_hex
from ctypes import (cdll, c_void_p, POINTER, c_ulong, c_char, c_uint32, pointer, byref, create_string_buffer, c_wchar,
                    cast, c_char_p)
from .const import SCardConstants

logger = logging.getLogger(__name__)

def i2b(i):
    return i.to_bytes(1, byteorder="little")

def b2i(data):
    return int(b2a_hex(data), 16)

class SCARDCONTEXT(c_uint32):
    pass

class SCARDHANDLE(c_uint32):
    pass

class SCardError(Exception):
    pass

class SCard(SCardConstants):
    def __init__(self):
        if platform.system() == "Darwin":
            lib_name = "PCSC.framework/PCSC"
        elif platform.system() == "Linux":
            lib_name = "libpcsclite.so"
        elif platform.system() == "Windows":
            lib_name = "winscard.dll"
        self.pcsc = cdll.LoadLibrary(lib_name)

    def __call__(self, method, *args):
        logger.debug(method + str(args))
        status = getattr(self.pcsc, method)(*args)
        if status != self.SCardStatus.S_SUCCESS.value:
            raise SCardError(self.SCardStatus(status))
        return status

    def EstablishContext(self, dwScope: SCardConstants.Scope,
                         pvReserved1: c_void_p,
                         pvReserved2: c_void_p,
                         phContext: SCARDCONTEXT) -> SCardConstants.SCardStatus:
        """
        The SCardEstablishContext function establishes the resource manager context (the scope) within which database
        operations are performed.
        """
        return self("SCardEstablishContext", dwScope, pvReserved1, pvReserved2, phContext)

    def ReleaseContext(self, hContext: SCARDCONTEXT) -> SCardConstants.SCardStatus:
        """
        The SCardReleaseContext function closes an established resource manager context, freeing any resources allocated
        under that context, including SCARDHANDLE objects and memory allocated using the SCARD_AUTOALLOCATE length
        designator.
        """
        return self("SCardReleaseContext", hContext)

    def IsValidContext(self, hContext: SCARDCONTEXT) -> SCardConstants.SCardStatus:
        """
        The SCardIsValidContext function determines whether a smart card context handle is valid.
        """
        return self("SCardIsValidContext", hContext)

    def SetTimeout(self, hContext: SCARDCONTEXT, dwTimeout: int) -> SCardConstants.SCardStatus:
        return self("SCardSetTimeout", hContext, dwTimeout)

    def Connect(self, hContext: SCARDCONTEXT,
                szReader: str, # const char *szReader, The name of the reader that contains the target card.
                dwShareMode: SCardConstants.ShareMode,
                dwPreferredProtocols: SCardConstants.Protocol,
                phCard: SCARDHANDLE,
                pdwActiveProtocol: SCardConstants.Protocol) -> SCardConstants.SCardStatus:
        """
        The SCardConnect function establishes a connection (using a specific resource manager context) between the calling
        application and a smart card contained by a specific reader. If no card exists in the specified reader, an error is
        returned.
        """
        return self("SCardConnect", hContext, szReader, dwShareMode, dwPreferredProtocols, phCard, pdwActiveProtocol)

    def Reconnect(self, hCard: SCARDHANDLE,
                  dwShareMode: SCardConstants.ShareMode,
                  dwPreferredProtocols: SCardConstants.Protocol,
                  dwInitialization: SCardConstants.Disposition,
                  pdwActiveProtocol: SCardConstants.Protocol) -> SCardConstants.SCardStatus:
        """
        The SCardReconnect function reestablishes an existing connection between the calling application and a smart
        card. This function moves a card handle from direct access to general access, or acknowledges and clears an
        error condition that is preventing further access to the card.
        """
        return self("SCardConnect", hCard, dwShareMode, dwPreferredProtocols, dwInitialization, pdwActiveProtocol)

    def Disconnect(self, hCard: SCARDHANDLE, dwDisposition: SCardConstants.Disposition) -> SCardConstants.SCardStatus:
        """
        The SCardDisconnect function terminates a connection previously opened between the calling application and a smart
        card in the target reader.
        """
        return self("SCardDisconnect", hCard, dwDisposition)

    def BeginTransaction(self, hCard: SCARDHANDLE) -> SCardConstants.SCardStatus:
        """
        The SCardBeginTransaction function starts a transaction.

        The function waits for the completion of all other transactions before it begins. After the transaction starts,
        all other applications are blocked from accessing the smart card while the transaction is in progress.
        """
        return self("SCardBeginTransaction", hCard)

    def EndTransaction(self, hCard: SCARDHANDLE, dwDisposition: SCardConstants.Disposition) -> SCardConstants.SCardStatus:
        """
        The SCardEndTransaction function completes a previously declared transaction, allowing other applications to resume
        interactions with the card.
        """
        return self("SCardEndTransaction", hCard, dwDisposition)

    def CancelTransaction(self, hCard: SCARDHANDLE) -> SCardConstants.SCardStatus:
        raise NotImplementedError()

    def Status(self, hCard: SCARDHANDLE,
               mszReaderNames: str,
               pcchReaderLen,
               pdwState: SCardConstants.CardState,
               pdwProtocol: SCardConstants.Protocol,
               pbAtr,
               pcbAtrLen) -> SCardConstants.SCardStatus:
        """
        The SCardStatus function provides the current status of a smart card in a reader. You can call it any time after a
        successful call to SCardConnect and before a successful call to SCardDisconnect. It does not affect the state of
        the reader or reader driver.
        """
        return self("SCardStatus", hCard, mszReaderNames, pcchReaderLen, pdwState, pdwProtocol, pbAtr, pcbAtrLen)

    def GetStatusChange(self, hContext: SCARDCONTEXT,
                        dwTimeout,
                        rgReaderStates: SCardConstants.ReaderState,
                        cReaders) -> SCardConstants.SCardStatus:
        return self("SCardGetStatusChange", hContext, dwTimeout, rgReaderStates, cReaders)

    def Control(self, hCard: SCARDHANDLE,
                dwControlCode,
                pbSendBuffer,
                cbSendLength,
                pbRecvBuffer,
                cbRecvLength,
                lpBytesReturned) -> SCardConstants.SCardStatus:
        """
        The SCardControl function gives you direct control of the reader. You can call it any time after a successful 
        call to SCardConnect and before a successful call to SCardDisconnect. The effect on the state of the reader
        depends on the control code.
        """
        return self("SCardControl", hCard, dwControlCode, pbSendBuffer, cbSendLength, pbRecvBuffer, cbRecvLength,
                    lpBytesReturned)

    def Transmit(self, hCard: SCARDHANDLE,
                 pioSendPci,
                 pbSendBuffer,
                 cbSendLength,
                 pioRecvPci,
                 pbRecvBuffer,
                 pcbRecvLength) -> SCardConstants.SCardStatus:
        """
        The SCardTransmit function sends a service request to the smart card and expects to receive data back from the
        card.
        """
        return self("SCardTransmit", hCard, pioSendPci, pbSendBuffer, cbSendLength, pioRecvPci, pbRecvBuffer,
                    pcbRecvLength)

    def ListReaderGroups(self, hContext: SCARDCONTEXT,
                         mszGroups, #char *mszGroups,
                         pcchGroups #uint32_t *pcchGroups
    ) -> SCardConstants.SCardStatus:
        return self("SCardListReaderGroups", hContext, mszGroups, pcchGroups)

    def ListReaders(self, hContext: SCARDCONTEXT, mszGroups, mszReaders, pcchReaders) -> SCardConstants.SCardStatus:
        """
        The SCardListReaders function provides the list of readers within a set of named reader groups, eliminating
        duplicates.

        The caller supplies a list of reader groups, and receives the list of readers within the named groups.
        Unrecognized group names are ignored. This function only returns readers within the named groups that
        are currently attached to the system and available for use.
        """
        return self("SCardListReaders", hContext, mszGroups, mszReaders, pcchReaders)

    def Cancel(self, hContext: SCARDCONTEXT) -> SCardConstants.SCardStatus:
        """
        The SCardCancel function terminates all outstanding actions within a specific resource manager context.

        The only requests that you can cancel are those that require waiting for external action by the smart card or
        user. Any such outstanding action requests will terminate with a status indication that the action was
        canceled. This is especially useful to force outstanding SCardGetStatusChange calls to terminate.
        """
        return self("SCardCancel", hContext)

    def GetAttrib(self, hCard: SCARDHANDLE, dwAttrId, pbAttr, pcbAttrLen) -> SCardConstants.SCardStatus:
        """
        The SCardGetAttrib function retrieves the current reader attributes for the given handle. It does not affect the
        state of the reader, driver, or card.
        """
        return self("SCardGetAttrib", hCard, dwAttrId, pbAttr, pcbAttrLen)

    def SetAttrib(self, hCard: SCARDHANDLE, dwAttrId, pbAttr, cbAttrLen) -> SCardConstants.SCardStatus:
        """
        The SCardSetAttrib function sets the given reader attribute for the given handle. It does not affect the state of
        the reader, reader driver, or smart card. Not all attributes are supported by all readers (nor can they be set
        at all times) as many of the attributes are under direct control of the transport protocol.
        """
        return self("SCardSetAttrib", hCard, dwAttrId, pbAttr, cbAttrLen)


class SCardManager(SCard):
    def __init__(self):
        SCard.__init__(self)
        self.ctx = SCARDCONTEXT()
        self.handle = SCARDHANDLE()
        self.protocol = c_ulong()
        self.EstablishContext(dwScope=self.Scope.SYSTEM, pvReserved1=0, pvReserved2=0, phContext=byref(self.ctx))

    def _split_multi_string(self, ms):
        p = cast(ms, POINTER(c_char))
        return p[:len(ms)].split(b"\0")

    def __iter__(self):
        pcch_readers = c_uint32()
        self.ListReaders(hContext=self.ctx, mszGroups=0, mszReaders=0, pcchReaders=byref(pcch_readers))
        s = create_string_buffer(b"\0" * pcch_readers.value)
        self.ListReaders(hContext=self.ctx, mszGroups=0, mszReaders=s, pcchReaders=byref(pcch_readers))
        for reader in self._split_multi_string(s):
            if reader:
                yield reader.decode()

    def __enter__(self):
        self.Connect(hContext=self.ctx,
                     szReader=c_char_p(list(self)[0].encode()),
                     dwShareMode=SCardConstants.ShareMode.SHARED,
                     dwPreferredProtocols=SCardConstants.Protocol.T0 | SCardConstants.Protocol.T1,
                     phCard=byref(self.handle),
                     pdwActiveProtocol=byref(self.protocol))

    def __exit__(self, exc_type, exc_value, traceback):
        self.Disconnect(hCard=self.handle, dwDisposition=SCardConstants.Disposition.LEAVE_CARD)

    def _get_send_pci(self):
        if self.protocol.value == self.Protocol.T0:
            return self.pcsc.g_rgSCardT0Pci
        elif self.protocol.value == self.Protocol.T1:
            return self.pcsc.g_rgSCardT1Pci

    def send_apdu(self, cla, ins, p1, p2, data):
        send_buf = create_string_buffer(i2b(cla) + i2b(ins) + i2b(p1) + i2b(p2) + i2b(len(data)) + data)
        recv_buf = create_string_buffer(b"\0" * self.MAX_BUFFER_SIZE_EXTENDED)
        recv_len = c_ulong(len(recv_buf))
        self.Transmit(hCard=self.handle,
                      pioSendPci=self._get_send_pci(),
                      pbSendBuffer=send_buf,
                      cbSendLength=len(send_buf),
                      pioRecvPci=0,
                      pbRecvBuffer=recv_buf,
                      pcbRecvLength=byref(recv_len))
        return recv_buf.raw[:recv_len.value]



#!/usr/bin/env python

import os, sys, unittest, json, collections

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))  # noqa

from exile import YKOATH, SCardManager

class TestExile(unittest.TestCase):
    def test_exile(self):
        ykoath = YKOATH(SCardManager())

if __name__ == '__main__':
    unittest.main()



from enum import Enum


class SCardConstants:
    """
    https://docs.microsoft.com/en-us/windows/desktop/api/winscard
    """
    MAX_BUFFER_SIZE = 264
    """Maximum Tx/Rx Buffer for short APDU"""
    MAX_BUFFER_SIZE_EXTENDED = 4 + 3 + (1 << 16) + 3
    """enhanced (64K + APDU + Lc + Le) Tx/Rx Buffer"""
    MAX_ATR_SIZE = 33
    MAX_READERNAME = 52
    class SCardStatus(Enum):
        S_SUCCESS = 0x00000000
        """No error was encountered."""
        F_INTERNAL_ERROR = 0x80100001
        """An internal consistency check failed."""
        E_CANCELLED = 0x80100002
        """The action was cancelled by an SCardCancel request."""
        E_INVALID_HANDLE = 0x80100003
        """The supplied handle was invalid."""
        E_INVALID_PARAMETER = 0x80100004
        """One or more of the supplied parameters could not be properly interpreted."""
        E_INVALID_TARGET = 0x80100005
        """Registry startup information is missing or invalid."""
        E_NO_MEMORY = 0x80100006
        """Not enough memory available to complete this command."""
        F_WAITED_TOO_LONG = 0x80100007
        """An internal consistency timer has expired."""
        E_INSUFFICIENT_BUFFER = 0x80100008
        """The data buffer to receive returned data is too small for the returned data."""
        E_UNKNOWN_READER = 0x80100009
        """The specified reader name is not recognized."""
        E_TIMEOUT = 0x8010000A
        """The user-specified timeout value has expired."""
        E_SHARING_VIOLATION = 0x8010000B
        """The smart card cannot be accessed because of other connections outstanding."""
        E_NO_SMARTCARD = 0x8010000C
        """The operation requires a Smart Card, but no Smart Card is currently in the device."""
        E_UNKNOWN_CARD = 0x8010000D
        """The specified smart card name is not recognized."""
        E_CANT_DISPOSE = 0x8010000E
        """The system could not dispose of the media in the requested manner."""
        E_PROTO_MISMATCH = 0x8010000F
        """The requested protocols are incompatible with the protocol currently in use with the smart card."""
        E_NOT_READY = 0x80100010
        """The reader or smart card is not ready to accept commands."""
        E_INVALID_VALUE = 0x80100011
        """One or more of the supplied parameters values could not be properly interpreted."""
        E_SYSTEM_CANCELLED = 0x80100012
        """The action was cancelled by the system, presumably to log off or shut down."""
        F_COMM_ERROR = 0x80100013
        """An internal communications error has been detected."""
        F_UNKNOWN_ERROR = 0x80100014
        """An internal error has been detected, but the source is unknown."""
        E_INVALID_ATR = 0x80100015
        """An ATR obtained from the registry is not a valid ATR string."""
        E_NOT_TRANSACTED = 0x80100016
        """An attempt was made to end a non-existent transaction."""
        E_READER_UNAVAILABLE = 0x80100017
        """The specified reader is not currently available for use."""
        P_SHUTDOWN = 0x80100018
        """The operation has been aborted to allow the server application to exit."""
        E_PCI_TOO_SMALL = 0x80100019
        """The PCI Receive buffer was too small."""
        E_READER_UNSUPPORTED = 0x8010001A
        """The reader driver does not meet minimal requirements for support."""
        E_DUPLICATE_READER = 0x8010001B
        """The reader driver did not produce a unique reader name."""
        E_CARD_UNSUPPORTED = 0x8010001C
        """The smart card does not meet minimal requirements for support."""
        E_NO_SERVICE = 0x8010001D
        """The Smart card resource manager is not running."""
        E_SERVICE_STOPPED = 0x8010001E
        """The Smart card resource manager has shut down."""
        E_UNEXPECTED = 0x8010001F
        """An unexpected card error has occurred."""
        E_ICC_INSTALLATION = 0x80100020
        """No primary provider can be found for the smart card."""
        E_ICC_CREATEORDER = 0x80100021
        """The requested order of object creation is not supported."""
        E_UNSUPPORTED_FEATURE = 0x80100022
        """This smart card does not support the requested feature."""
        E_DIR_NOT_FOUND = 0x80100023
        """The identified directory does not exist in the smart card."""
        E_FILE_NOT_FOUND = 0x80100024
        """The identified file does not exist in the smart card."""
        E_NO_DIR = 0x80100025
        """The supplied path does not represent a smart card directory."""
        E_NO_FILE = 0x80100026
        """The supplied path does not represent a smart card file."""
        E_NO_ACCESS = 0x80100027
        """Access is denied to this file."""
        E_WRITE_TOO_MANY = 0x80100028
        """The smart card does not have enough memory to store the information."""
        E_BAD_SEEK = 0x80100029
        """There was an error trying to set the smart card file object pointer."""
        E_INVALID_CHV = 0x8010002A
        """The supplied PIN is incorrect."""
        E_UNKNOWN_RES_MNG = 0x8010002B
        """An unrecognized error code was returned from a layered component."""
        E_NO_SUCH_CERTIFICATE = 0x8010002C
        """The requested certificate does not exist."""
        E_CERTIFICATE_UNAVAILABLE = 0x8010002D
        """The requested certificate could not be obtained."""
        E_NO_READERS_AVAILABLE = 0x8010002E
        """Cannot find a smart card reader."""
        E_COMM_DATA_LOST = 0x8010002F
        """A communications error with the smart card has been detected. Retry the operation."""
        E_NO_KEY_CONTAINER = 0x80100030
        """The requested key container does not exist on the smart card."""
        E_SERVER_TOO_BUSY = 0x80100031
        """The Smart Card Resource Manager is too busy to complete this operation."""

        W_UNSUPPORTED_CARD = 0x80100065
        """The reader cannot communicate with the card, due to ATR string configuration conflicts."""
        W_UNRESPONSIVE_CARD = 0x80100066
        """The smart card is not responding to a reset."""
        W_UNPOWERED_CARD = 0x80100067
        """Power has been removed from the smart card, so that further communication is not possible."""
        W_RESET_CARD = 0x80100068
        """The smart card has been reset, so any shared state information is invalid."""
        W_REMOVED_CARD = 0x80100069
        """The smart card has been removed, so further communication is not possible."""

        W_SECURITY_VIOLATION = 0x8010006A
        """Access was denied because of a security violation."""
        W_WRONG_CHV = 0x8010006B
        """The card cannot be accessed because the wrong PIN was presented."""
        W_CHV_BLOCKED = 0x8010006C
        """The card cannot be accessed because the maximum number of PIN entry attempts has been reached."""
        W_EOF = 0x8010006D
        """The end of the smart card file has been reached."""
        W_CANCELLED_BY_USER = 0x8010006E
        """The user pressed "Cancel" on a Smart Card Selection Dialog."""
        W_CARD_NOT_AUTHENTICATED = 0x8010006F
        """No PIN was presented to the smart card."""

    class Scope:
        USER = 0x0000
        """Scope in user space"""
        TERMINAL = 0x0001
        """Scope in terminal"""
        SYSTEM = 0x0002
        """Scope in system"""

    class Protocol:
        UNDEFINED = 0x0000
        """protocol not set"""
        UNSET = UNDEFINED
        T0 = 0x0001
        """T=0 active protocol."""
        T1 = 0x0002
        """T=1 active protocol."""
        RAW = 0x0004
        """Raw active protocol."""
        T15 = 0x0008
        """T=15 protocol."""
        ANY = T0 | T1
        """IFD determines prot."""

    class ShareMode:
        EXCLUSIVE = 0x0001
        """Exclusive mode only"""
        SHARED = 0x0002
        """Shared mode only"""
        DIRECT = 0x0003
        """Raw mode only"""

    class Disposition:
        LEAVE_CARD = 0x0000
        """Do nothing on close"""
        RESET_CARD = 0x0001
        """Reset on close"""
        UNPOWER_CARD = 0x0002
        """Power down on close"""
        EJECT_CARD = 0x0003
        """Eject on close"""

    class CardState:
        UNKNOWN = 0x0001
        """Unknown state"""
        ABSENT = 0x0002
        """Card is absent"""
        PRESENT = 0x0004
        """Card is present"""
        SWALLOWED = 0x0008
        """Card not powered"""
        POWERED = 0x0010
        """Card is powered"""
        NEGOTIABLE = 0x0020
        """Ready for PTS"""
        SPECIFIC = 0x0040
        """PTS has been set"""

    class ReaderState:
        UNAWARE = 0x0000
        """App wants status"""
        IGNORE = 0x0001
        """Ignore this reader"""
        CHANGED = 0x0002
        """State has changed"""
        UNKNOWN = 0x0004
        """Reader unknown"""
        UNAVAILABLE = 0x0008
        """Status unavailable"""
        EMPTY = 0x0010
        """Card removed"""
        PRESENT = 0x0020
        """Card inserted"""
        ATRMATCH = 0x0040
        """ATR matches card"""
        EXCLUSIVE = 0x0080
        """Exclusive Mode"""
        INUSE = 0x0100
        """Shared Mode"""
        MUTE = 0x0200
        """Unresponsive card"""
        UNPOWERED = 0x0400
        """Unpowered card"""



import struct
from enum import Enum
from ..scard import i2b


class YKOATHConstants:
    HMAC_MINIMUM_KEY_SIZE = 14

    class Tag:
        NAME = 0x71
        NAME_LIST = 0x72
        KEY = 0x73
        CHALLENGE = 0x74
        RESPONSE = 0x75
        TRUNCATED_RESPONSE = 0x76
        NO_RESPONSE = 0x77
        PROPERTY = 0x78
        VERSION = 0x79
        IMF = 0x7a
        ALGORITHM = 0x7b
        TOUCH = 0x7c

    class OATHType:
        HOTP = 0x10
        TOTP = 0x20

    class Properties:
        REQUIRE_TOUCH = 0x02

    class Instruction:
        SELECT = 0xa4
        PUT = 0x01
        DELETE = 0x02
        SET_CODE = 0x03
        RESET = 0x04
        LIST = 0xa1
        CALCULATE = 0xa2
        VALIDATE = 0xa3
        CALCULATE_ALL = 0xa4
        SEND_REMAINING = 0xa5

    class Algorithm:
        SHA1 = 0x01
        SHA256 = 0x02
        SHA512 = 0x03

    class Application:
        OTP = b'\xa0\x00\x00\x05\x27\x20\x01'
        MGR = b'\xa0\x00\x00\x05\x27\x47\x11\x17'
        OPGP = b'\xd2\x76\x00\x01\x24\x01'
        OATH = b'\xa0\x00\x00\x05\x27\x21\x01'
        PIV = b'\xa0\x00\x00\x03\x08'
        U2F = b'\xa0\x00\x00\x06\x47\x2f\x00\x01'

    class Error(Enum):
        SUCCESS = b'\x90\x00'
        NO_SPACE = b'\x6a\x84'
        NOT_FOUND = b'\x69\x84'
        AUTH_REQUIRED = b'x69\x82'
        WRONG_SYNTAX = b'\x6a\x80'
        GENERIC_ERROR = b'\x65\x81'
        MORE_DATA_AVAILABLE = b'\x61'

class YKOATHError(Exception):
    pass

class YKOATH(YKOATHConstants):
    """
    See https://developers.yubico.com/OATH/YKOATH_Protocol.html
    """
    def __init__(self, scm):
        self.scm = scm
        self.send_apdu(cla=0, ins=self.Instruction.SELECT, p1=0x04, p2=0, data=self.Application.OATH)

    def send_apdu(self, **kwargs):
        with self.scm:
            res = self.scm.send_apdu(**kwargs)
        if res[-2:] != self.Error.SUCCESS.value:
            raise YKOATHError(self.Error(res[-2:]))
        return res

    def put(self, credential_name, secret,
            oath_type=YKOATHConstants.OATHType.TOTP, algorithm=YKOATHConstants.Algorithm.SHA1, digits=6):
        secret_header = i2b(oath_type | algorithm) + i2b(digits)
        # secret = hmac_shorten_key(secret, algorithm)
        secret = secret.ljust(self.HMAC_MINIMUM_KEY_SIZE, b'\x00')
        data = i2b(self.Tag.NAME) + i2b(len(credential_name)) + credential_name.encode()
        data += i2b(self.Tag.KEY) + i2b(len(secret_header) + len(secret)) + secret_header + secret
        # data += bytearray([YKOATH.Tag.PROPERTY, YKOATH.Properties.REQUIRE_TOUCH])
        return self.send_apdu(cla=0, ins=self.Instruction.PUT, p1=0, p2=0, data=data)

    def delete(self, credential_name):
        data = i2b(self.Tag.NAME) + i2b(len(credential_name)) + credential_name.encode()
        return self.send_apdu(cla=0, ins=self.Instruction.DELETE, p1=0, p2=0, data=data)

    def reset(self):
        return self.send_apdu(cla=0, ins=self.Instruction.RESET, p1=0xde, p2=0xad, data=b"")

    def list(self):
        return self.send_apdu(cla=0, ins=self.Instruction.LIST, p1=0, p2=0, data=b"")

    def calculate(self, credential_name, challenge, want_truncated_response=True):
        if not isinstance(challenge, bytes):
            challenge = self.int_to_bytestring(challenge)
        data = i2b(self.Tag.NAME) + i2b(len(credential_name)) + credential_name.encode()
        data += i2b(self.Tag.CHALLENGE) + i2b(len(challenge)) + challenge
        p2 = 0x01 if want_truncated_response else 0
        res = self.send_apdu(cla=0, ins=self.Instruction.CALCULATE, p1=0, p2=p2, data=data)
        assert res[0] == self.Tag.TRUNCATED_RESPONSE if want_truncated_response else self.Tag.RESPONSE
        res_len, digits = res[1], res[2]
        if want_truncated_response:
            return str(struct.unpack('>I', res[3:3+res_len-1])[0]).zfill(digits)
        else:
            return res[3:3+res_len-1]

    def int_to_bytestring(self, i, padding=8):
        result = bytearray()
        while i != 0:
            result.append(i & 0xFF)
            i >>= 8
        return bytes(bytearray(reversed(result)).rjust(padding, b'\0'))
